import io
import os
import re
import sys
import urllib.request

import pandas as pd


def fetch(tracks, path, clobber=True):
    """Download MARSIS data from the PDS.

    Download MARSIS data from the PDS by searching the PDS
    index files to match user supplied file names.

    Parameters
    ----------
    tracks: list of str
        List of the MARSIS data files to download, each entry in the
        format "E_XXXXX_SS3_TRK_CMP_M"
    path: str
        Path to directory where the MARSIS data files will be downloaded to
    clobber: bool
        Whether to overwrite (re-download) files at path

    Returns
    -------
    files: list of str
        Tuple containing the paths of all files downloaded by this function

    Notes
    -----
    """

    # Only allow ss3_trk_cmp for now, raise error if anything different
    r = re.compile("e_[0-9]{5}_ss3_trk_cmp_m")
    for track in tracks:
        if r.match(track.lower()) is None:
            print(track)
            raise ValueError("Processor only supports E_SS3_TRK_CMP type")

    # Get index files from pds
    print("Downloading PDS index files...")

    baseurl = "https://pds-geosciences.wustl.edu/mex/"
    indxURLs = [
        baseurl + "mex-m-marsis-2-edr-v2/mexme_1001/index/index.tab",
        baseurl + "mex-m-marsis-2-edr-ext1-v2/mexme_1002/index/index.tab",
        baseurl + "mex-m-marsis-2-edr-ext2-v1/mexme_1003/index/index.tab",
        baseurl + "mex-m-marsis-2-edr-ext3-v1/mexme_1004/index/index.tab",
        baseurl + "mex-m-marsis-2-edr-ext4-v1/mexme_1005/index/index.tab",
        baseurl + "mex-m-marsis-2-edr-ext5-v1/mexme_1006/index/index.tab",
        baseurl + "mex-m-marsis-2-edr-ext6-v1/mexme_1007/index/index.tab",
        baseurl + "mex-m-marsis-2-edr-ext7-v1/mexme_1008/index/index.tab",
    ]

    index = ""

    try:
        for url in indxURLs:
            data = urllib.request.urlopen(url)
            subIndx = str(data.read(), "utf-8")
            subPath = url.split("/")[:6]
            subPath = "/".join(subPath) + "/"
            subIndx = subIndx.replace("DATA/", subPath + "DATA/").lower()
            index += subIndx
    except Exception as e:
        print(e, file=sys.stderr)
        print("Error downloading index file %s" % url, file=sys.stderr)

    # Create pandas dataframe
    print("Creating index dataframe...")

    cols = ["LBL_PATH", "ID", "TIME", "DS_ID", "RLS_ID", "REV_ID"]
    df = pd.read_csv(io.StringIO(index), sep=",", header=None, names=cols)

    # Download the data files
    print("Downloading data files...")

    files = [None] * len(tracks) * 3
    try:
        for track in tracks:
            track = track.lower()
            lbl = df[df["ID"] == track]["LBL_PATH"]
            if len(lbl) == 0:
                raise ValueError("No files found to match %s" % track)
            lbl = lbl.iloc[0]
            geo = lbl.replace(".lbl", "_g.dat")
            sci = lbl.replace(".lbl", "_f.dat")

            lbl_path = path + "/" + lbl.split("/")[-1]
            if clobber or not os.path.isfile(lbl_path):
                lbl_path, headers = urllib.request.urlretrieve(lbl, lbl_path)

            geo_path = path + "/" + geo.split("/")[-1]
            if clobber or not os.path.isfile(geo_path):
                geo_path, headers = urllib.request.urlretrieve(geo, geo_path)

            sci_path = path + "/" + sci.split("/")[-1]
            if clobber or not os.path.isfile(sci_path):
                sci_path, headers = urllib.request.urlretrieve(sci, sci_path)

            files.append(lbl_path)
            files.append(geo_path)
            files.append(sci_path)

    except Exception as e:
        print(e, file=sys.stderr)
        print("Error downloading track %s" % track, file=sys.stderr)

    return files



import os
import struct
import sys

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from ply import lex, yacc


class EDR:
    """Reader for experimental data record (EDR) files from the PDS.

    This object will ingest and store data from the EDR files, to be
    processed later. It only ingests data into a convenient structure,
    and de-compresses the science data, no processing occurs.
    """

    def __init__(self, lbl=None):
        """Create an EDR object.

        Create an EDR object, read in an EDR file if given.

        Parameters
        ----------
        lbl: str, optional
            Path to an EDR label file which must be in the same directory as
            the corresponding science and geometry files

        Returns
        -------
        EDR

        Notes
        -----
        """
        if lbl is None:
            return

        self.load(lbl)

    def load(self, lbl):
        """Load a set of EDR files.

        Read in the label, science, and geometry EDR files for a given
        observation.

        Parameters
        ----------
        lbl: str
            Path to an EDR label file which must be in the same directory as
            the corresponding science and geometry files

        Returns
        -------

        Notes
        -----
        """
        # Read label file
        self.lbld = self.parseLBL(lbl)

        # Science and aux file names
        sci = lbl.replace(".lbl", "_f.dat")
        geo = lbl.replace(".lbl", "_g.dat")

        # Read science file
        self.anc, self.ost, self.data = self.parseSci(sci, self.lbld)

        # Read geometry file
        self.geo = self.parseGeo(geo, self.lbld)

    def parseGeo(self, file, lbld):
        """Read in a geometry data file.

        Parameters
        ----------
        file: str
            Path to an geometry data file
        lbld: dict
            Dictionary containing data parsed from corresponding label
            file, created by the parseLBL method

        Returns
        -------

        Notes
        -----
        """
        # Set up geometry data frame

        rec_t = np.dtype(
            [
                ("SCET_FRAME_WHOLE", ">u4"),
                ("SCET_FRAME_FRAC", ">u2"),
                ("GEOMETRY_EPHEMERIS_TIME", ">f8"),
                ("GEOMETRY_EPOCH", "V23"),
                ("MARS_SOLAR_LONGITUDE", ">f8"),
                ("MARS_SUN_DISTANCE", ">f8"),
                ("ORBIT_NUMBER", ">u4"),
                ("TARGET_NAME", "V6"),
                ("TARGET_SC_POSITION_VECTOR", ">f8", 3),
                ("SPACECRAFT_ALTITUDE", ">f8"),
                ("SUB_SC_LONGITUDE", ">f8"),
                ("SUB_SC_LATITUDE", ">f8"),
                ("TARGET_SC_VELOCITY_VECTOR", ">f8", 3),
                ("TARGET_SC_RADIAL_VELOCITY", ">f8"),
                ("TARGET_SC_TANG_VELOCITY", ">f8"),
                ("LOCAL_TRUE_SOLAR_TIME", ">f8"),
                ("SOLAR_ZENITH_ANGLE", ">f8"),
                ("DIPOLE_UNIT_VECTOR", ">f8", 3),
                ("MONOPOLE_UNIT_VECTOR", ">f8", 3),
            ]
        )

        geoData = np.fromfile(file, dtype=rec_t)

        return geoData

    def decompress(self, trace, exp):
        sign = (-1) ** (trace >> 7)  # Sign bit
        mantissa = 1 + ((trace & 0x7F) / (2.0 ** 7))
        # mantissa = trace & 0x7F
        trace = sign * mantissa * (2 ** (exp - 127))
        return trace

    def deagc(self, data, agc):
        # Take in an array of marsis data
        # and a vector of AGC settings,
        # then correct for agc
        agc = agc & 0x07  # Only last three bits matter
        agc = agc * 4 + 2  # Gain in dB, per Orosei
        data = data * 10 ** (agc / 20)[np.newaxis, :]
        return data

    def parseSci(self, file, lbld):
        # Set up ancillary data dataframe

        rec_t = np.dtype(
            [
                ("SCET_STAR_WHOLE", ">u4"),
                ("SCET_STAR_FRAC", ">u2"),
                ("OST_LINE_NUMBER", ">u2"),
                ("OST_LINE", "V12"),
                ("FRAME_NUMBER", ">u2"),
                ("ANCILLARY_DATA_HEADER", "V6"),
                ("FIRST_PRI_OF_FRAME", ">u4"),
                ("SCET_FRAME_WHOLE", ">u4"),
                ("SCET_FRAME_FRAC", ">u2"),
                ("SCET_PERICENTER_WHOLE", ">u4"),
                ("SCET_PERICENTER_FRAC", ">u2"),
                ("SCET_PAR_WHOLE", ">u4"),
                ("SCET_PAR_FRAC", ">u2"),
                ("H_SCET_PAR", ">f4"),
                ("VT_SCET_PAR", ">f4"),
                ("VR_SCET_PAR", ">f4"),
                ("N_0", ">u4"),
                ("DELTA_S_MIN", ">f4"),
                ("NB_MIN", ">u2"),
                ("M_OCOG", ">f4", 2),
                ("INDEX_OCOG", ">u2", 2),
                ("TRK_THRESHOLD", ">f4", 2),
                ("INI_IND_TRK_THRESHOLD", ">u2", 2),
                ("LAST_IND_TRK_THRESHOLD", ">u2", 2),
                ("INI_IND_FSRM", ">u2", 2),
                ("LAST_IND_FSRM", ">u2", 2),
                ("SPARE_4", ">u4", 3),
                ("DELTA_S_SCET_PAR", ">f4"),
                ("NB_SCET_PAR", ">u2"),
                ("NA_SCET_PAR", ">u2", 2),
                ("A2_INI_CM", ">f4", 2),
                ("A2_OPT", ">f4", 2),
                ("REF_CA_OPT", ">f4", 2),
                ("DELTA_T", ">u2", 2),
                ("SF", ">f4", 2),
                ("I_C", ">u2", 2),
                ("AGC_SA_FOR_NEXT_FRAME", ">f4", 2),
                ("AGC_SA_LEVELS_CURRENT_FRAME", ">u1", 2),
                ("RX_TRIG_SA_FOR_NEXT_FRAME", ">u2", 2),
                ("RX_TRIG_SA_PROGR", ">u2", 2),
                ("INI_IND_OCOG", ">u2"),
                ("LAST_IND_OCOG", ">u2"),
                ("OCOG", ">f4", 2),
                ("A", ">f4", 2),
                ("C_LOL", ">i2", 2),
                ("SPARE_5", ">u2", 3),
                ("MAX_RE_EXP_MINUS1_F1_DIP", ">u1"),
                ("MAX_IM_EXP_MINUS1_F1_DIP", ">u1"),
                ("MAX_RE_EXP_ZERO_F1_DIP", ">u1"),
                ("MAX_IM_EXP_ZERO_F1_DIP", ">u1"),
                ("MAX_RE_EXP_PLUS1_F1_DIP", ">u1"),
                ("MAX_IM_EXP_PLUS1_F1_DIP", ">u1"),
                ("MAX_RE_EXP_MINUS1_F2_DIP", ">u1"),
                ("MAX_IM_EXP_MINUS1_F2_DIP", ">u1"),
                ("MAX_RE_EXP_ZERO_F2_DIP", ">u1"),
                ("MAX_IM_EXP_ZERO_F2_DIP", ">u1"),
                ("MAX_RE_EXP_PLUS1_F2_DIP", ">u1"),
                ("MAX_IM_EXP_PLUS1_F2_DIP", ">u1"),
                ("SPARE_6", ">u1", 8),
                ("AGC_PIS_PT_VALUE", ">f4", 2),
                ("AGC_PIS_LEVELS", ">u1", 2),
                ("K_PIM", ">u1"),
                ("PIS_MAX_DATA_EXP", ">u1", 2),
                ("PROCESSING_PRF", ">f4"),
                ("SPARE_7", ">u1"),
                ("REAL_ECHO_MINUS1_F1_DIP", ">u1", 512),
                ("IMAG_ECHO_MINUS1_F1_DIP", ">u1", 512),
                ("REAL_ECHO_ZERO_F1_DIP", ">u1", 512),
                ("IMAG_ECHO_ZERO_F1_DIP", ">u1", 512),
                ("REAL_ECHO_PLUS1_F1_DIP", ">u1", 512),
                ("IMAG_ECHO_PLUS1_F1_DIP", ">u1", 512),
                ("REAL_ECHO_MINUS1_F2_DIP", ">u1", 512),
                ("IMAG_ECHO_MINUS1_F2_DIP", ">u1", 512),
                ("REAL_ECHO_ZERO_F2_DIP", ">u1", 512),
                ("IMAG_ECHO_ZERO_F2_DIP", ">u1", 512),
                ("REAL_ECHO_PLUS1_F2_DIP", ">u1", 512),
                ("IMAG_ECHO_PLUS1_F2_DIP", ">u1", 512),
                ("PIS_F1", ">i2", 128),
                ("PIS_F2", ">i2", 128),
            ]
        )

        telTab = np.fromfile(file, dtype=rec_t)

        # Decode OST line bit fields - this is incomplete
        df = pd.DataFrame()
        ost = telTab["OST_LINE"]
        ost = np.array(ost.tolist())  # weird but it works to get from void to bytes_
        df["SPARE_0"] = np.vectorize(lambda s: s[0])(ost)
        df["MODE_DURATION"] = np.vectorize(
            lambda s: np.frombuffer(s[0:4], dtype=">u4") & 0x00FFFFFF
        )(ost)
        df["SPARE_1"] = np.vectorize(
            lambda s: np.frombuffer(s[4:5], dtype=">u1") & 0xC0
        )(ost)
        df["MODE_SELECTION"] = np.vectorize(
            lambda s: np.frombuffer(s[4:5], dtype=">u1") >> 2 & 0x0F
        )(ost)
        df["DCG_CONFIGURATION_F1"] = np.vectorize(
            lambda s: np.frombuffer(s[4:5], dtype=">u1") & 0x03
        )(ost)
        df["DCG_CONFIGURATION_F2"] = np.vectorize(
            lambda s: np.frombuffer(s[5:6], dtype=">u1") >> 6
        )(ost)

        # Decompress data and make radargrams
        moded = {
            "SS3_TRK_CMP": [
                "MINUS1_F1",
                "ZERO_F1",
                "PLUS1_F1",
                "MINUS1_F2",
                "ZERO_F2",
                "PLUS1_F2",
            ]
        }

        mode = lbld["INSTRUMENT_MODE_ID"].replace('"', "")

        if mode not in moded.keys():
            print("Unhandled mode, exiting")
            print(mode)
            sys.exit()

        datad = {}

        for rg in moded[mode]:
            block = np.zeros((512, len(telTab)), dtype=np.complex64)
            for i in range(len(telTab)):
                expIM = telTab["MAX_IM_EXP_" + rg + "_DIP"][i]
                expRE = telTab["MAX_RE_EXP_" + rg + "_DIP"][i]
                trIM = telTab["IMAG_ECHO_" + rg + "_DIP"][i]
                trRE = telTab["REAL_ECHO_" + rg + "_DIP"][i]

                trRE = self.decompress(trRE, expRE)
                trIM = self.decompress(trIM, expIM)

                trace = trRE + 1j * trIM

                band = int(rg.split("_")[1][1])

                block[:, i] = trace

            if "F1" in rg:
                block = self.deagc(block, telTab["AGC_SA_LEVELS_CURRENT_FRAME"][:, 0])
            elif "F2" in rg:
                block = self.deagc(block, telTab["AGC_SA_LEVELS_CURRENT_FRAME"][:, 1])

            datad[rg] = block

        return telTab, df, datad

    def buildDict(self, pdata, i):
        dd = {}

        while i < len(pdata):
            key, val = pdata[i]
            if key == "OBJECT":
                c = 0
                name = val + str(c)
                while name in dd.keys():
                    c += 1
                    name = val + str(c)

                i, dd[name] = self.buildDict(pdata, i + 1)
                continue

            if key == "END_OBJECT":
                return i + 1, dd

            dd[key] = val
            i += 1

        return dd

    def parseLBL(self, lbl):
        # Parse the label file with lex and yacc
        # Heavily based on https://github.com/mkelley/pds3

        # lexer def ###

        tokens = [
            "DSID",
            "WORD",
            "STRING",
            "COMMENT",
            "POINTER",
            "DATE",
            "INT",
            "REAL",
            "UNIT",
            "END",
        ]

        literals = ["(", ")", ",", "=", "{", "}"]

        def t_DSID(t):
            r"MEX-M-MARSIS-2-EDR(-EXT[0-9])?-V[0-9].0"  # Dataset ID
            return t

        def t_WORD(t):
            r"[A-Z][A-Z0-9:_]+"
            if t.value == "END":
                t.type = "END"
            return t

        t_STRING = r'"[^"]+"'

        def t_COMMENT(t):
            r"/\*.+\*/"
            pass

        t_POINTER = r"\^[A-Z0-9_]+"
        t_DATE = r"[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}(.[0-9]{3})?"
        t_INT = r"[+-]?[0-9]+"
        t_REAL = r"[+-]?[0-9]+\.[0-9]+([Ee][+-]?[0-9]+)?"
        t_UNIT = r"<[\w*^\-/]+>"
        t_ignore = " \t\r\n"

        def t_error(t):
            print("Illegal character '%s'" % t.value[0])
            t.lexer.skip(1)

        lexer = lex.lex()

        # ## parser def ## #

        def p_label(p):
            """label : record
            | label record
            | label END"""
            if len(p) == 2:
                # record
                p[0] = [p[1]]
            elif p[2] == "END":
                # label END
                p[0] = p[1]
            else:
                # label record
                p[0] = p[1] + [p[2]]

        def p_record(p):
            """record : WORD '=' value
            | POINTER '=' INT
            | POINTER '=' STRING
            | POINTER '=' '(' STRING ',' INT ')'"""
            p[0] = (p[1], p[3])

        def p_value(p):
            """value : STRING
            | DATE
            | WORD
            | DSID
            | number
            | number UNIT
            | sequence"""
            # Just chuck the units for now
            p[0] = p[1]

        def p_number(p):
            """number : INT
            | REAL"""
            p[0] = p[1]

        def p_sequence(p):
            """sequence : '(' value ')'
            | '(' sequence_values ')'
            | '{' value '}'
            | '{' sequence_values '}'"""
            p[0] = p[2]

        def p_sequence_values(p):
            """sequence_values : value ','
            | sequence_values value ','
            | sequence_values value"""
            if p[2] == ",":
                p[0] = [p[1]]
            else:
                p[0] = p[1] + [p[2]]

        def p_error(p):
            if p:
                print("Syntax error at '%s'" % p.value)
            else:
                print("Syntax error at EOF")

        parser = yacc.yacc()

        # ## parse the label ## #

        fd = open(lbl, "r")
        data = fd.read()
        fd.close()

        result = parser.parse(data, lexer=lexer, debug=False)

        return self.buildDict(result, 0)



import numpy as np
from tqdm import tqdm


def arangeT(start, stop, fs):
    # Function to generate set of
    # Args are start time, stop time, sampling frequency
    # Generates times within the closed interval [start, stop] at 1/fs spacing
    # Double precision floating point

    tTot = stop - start
    nsamp = int(tTot * fs)
    seq = np.arange(nsamp, dtype=np.double)
    seq = seq / fs + start

    return seq


def psiGridSearch(TRACE, sim, tshift, band, dly, dlyBound, steps, plot=False):
    # Find min/max psi bounds based on delay
    dlyMax = dly + dlyBound
    dlyMin = max(0, dly - dlyBound)

    psi1C = {3: 3.04e-8, 4: 1.7e-8, 5: 1.08e-8}
    psi2C = {3: 2.83e-7, 4: 8.75e-8, 5: 3.54e-8}
    psi3C = {3: 1.69e-6, 4: 2.85e-7, 5: 7.30e-8}
    key = int(band / 1e6)
    psi1Max = dlyMax / psi1C[key]
    psi2Max = dlyMax / psi2C[key]
    psi3Max = dlyMax / psi3C[key]

    # Constrained grid search by dlyBound samples around dly
    # print(psi1Min, psi2Min, psi3Min)
    psi1 = np.linspace(0, psi1Max, steps)
    psi2 = np.linspace(0, psi2Max, steps)
    psi3 = np.linspace(0, psi3Max, steps)

    res = np.zeros((steps, steps, steps))
    for i in range(len(psi1)):
        for j in range(len(psi2)):
            for k in range(len(psi3)):
                delay = totalDelay([psi1[i], psi2[j], psi3[k]], band)
                if np.abs(delay - dly) > dlyBound:
                    res[i, j, k] = 0
                else:
                    res[i, j, k] = obj_func(
                        [psi1[i], psi2[j], psi3[[k]]], TRACE, tshift, band, sim
                    )

    foc = np.where(res == res.min())
    if plot:
        plt.figure()
        img = res[:, :, foc[2]].reshape((steps, steps))
        img = np.log(np.abs(img))
        plt.imshow(img, aspect="auto")
        plt.plot([foc[1]], [foc[0]], "r.")
    op1 = psi1[foc[0]]
    op2 = psi2[foc[1]]
    op3 = psi3[foc[2]]
    return [op1, op2, op3]


def totalDelay(psis, band):
    # Total delay in samples
    key = int(band / 1e6)
    coeffs = {
        3: [3.04e-8, 2.83e-7, 1.69e-6],
        4: [1.7e-8, 8.75e-8, 2.85e-7],
        5: [1.08e-8, 3.54e-8, 7.30e-8],
    }
    coeff = coeffs[key]
    return coeff[0] * psis[0] + coeff[1] * psis[1] + coeff[2] * psis[2]


def obj_func(psis, TRACE, ttrig, band, sim):
    pc = pc_phase(psis, TRACE, ttrig, band)

    num = np.sum((np.abs(pc) ** 2) * (sim ** 2))
    srf = np.argmax(sim)
    denom = np.sum(np.abs(pc[0:srf]) ** 2)

    cost = -num / (denom ** 2)

    if False:
        plt.figure()
        plt.plot(np.abs(pc))
        plt.plot(np.abs(sim) * (np.max(np.abs(pc)) / np.max(sim)))
        plt.title(psis)
        plt.figure()
        plt.plot(np.abs(pc[0:srf]))
        plt.title(psis)
        print(psis, num, denom)
    return cost


def pc_phase(psis, TRACE, ttrig, band):
    # psis - ionospheric distortion coeffs
    # TRACE - freq domain trace to pulse compress
    # ttrig - trigger time correction
    # f0 - carrier freq
    (psi1, psi2, psi3) = psis

    # Reference chirp
    fs = 1.4e6
    tlen = 250e-6
    f0 = -0.5e6
    m = 4.0e9
    t = arangeT(0, tlen, fs)  # time vector
    phi = np.pi * (2 * f0 * t + m * (t ** 2))  # phase
    chirp = -1 * np.sin(phi) + 1j * np.cos(phi)  # chirp
    # chirp = chirp*scipy.signal.windows.hann(len(chirp))

    # Zero pad data
    trace = np.fft.ifft(TRACE)
    # trace = np.append(trace, np.zeros(len(chirp)))

    # Zero pad chirp
    chirp = np.append(chirp, np.zeros(len(trace) - len(chirp)))

    # Shift data to baseband
    t = np.arange(len(chirp)) * 1.0 / fs
    bb = np.exp(2 * np.pi * 1j * -0.7e6 * t)
    trace = trace * bb

    # FFT
    CHIRP = np.fft.fft(chirp)
    TRACE = np.fft.fft(trace)

    # Pulse compress
    w = np.fft.fftfreq(len(chirp), d=1.0 / fs) * 2 * np.pi  # Angular frequency
    f = (np.fft.fftfreq(len(chirp), d=1.0 / fs) + band) / 1e6  # Frequency in MHz

    c = 299792458  # speed of light m/s
    PHASE = np.exp(
        ((-1j * 2 * np.pi) / c)
        * (
            (((8.98 ** 2) * psi1) / f)
            + (((8.98 ** 4) * psi2) / (3 * (f ** 3)))
            + (((8.98 ** 6) * psi3) / (8 * (f ** 5)))
        )
    )
    pc = np.fft.ifft(TRACE * np.conj(CHIRP) * np.exp(-1j * w * ttrig) * PHASE)
    return pc


def pc_clutter(DATA, sim, ttrig, dcg_config, psis=None):
    findPsis = False
    if psis is None:
        findPsis = True
        psi_t = np.dtype(
            [("psi1", np.float32), ("psi2", np.float32), ("psi3", np.float32)]
        )
        psis = np.empty(DATA.shape[1], dtype=psi_t)

    bands = [1.8e6, 3e6, 4e6, 5e6]
    pc = np.zeros(DATA.shape, dtype=np.float32)
    for i in tqdm(range(DATA.shape[1])):
        band = bands[dcg_config[i]]

        datat = DATA[:, i]
        simt = sim[:, i]
        ttrigt = ttrig[i]

        if findPsis:
            # Grid search ionosphere
            psi = psiGridSearch(datat, simt, ttrigt, band, 0, 256, 10)
            dly = totalDelay(psi, band)[0]

            psi = psiGridSearch(datat, simt, ttrigt, band, dly, 50, 10)
            dly = totalDelay(psi, band)[0]

            psi = psiGridSearch(datat, simt, ttrigt, band, dly, 5, 20)
            dly = totalDelay(psi, band)[0]

            psi = psiGridSearch(datat, simt, ttrigt, band, dly, 1, 40)
            dly = totalDelay(psi, band)[0]

            psis[i] = tuple(psi)

        pc[:, i] = np.abs(pc_phase(psis[i], datat, ttrigt, band))

    if findPsis:
        return pc, psis
    else:
        return pc


def mcmichael(edr, sim):
    c = 299792458
    dt = 1.0 / 1.4e6

    sim = np.fromfile(sim, dtype=np.float32).reshape(edr.data["ZERO_F1"].shape)

    tshiftF1 = (
        edr.anc["RX_TRIG_SA_PROGR"][:, 0] * (1 / 2.8e6)
        - edr.geo["SPACECRAFT_ALTITUDE"] * 2e3 / c
        + 256 * dt
    )
    tshiftF2 = (
        edr.anc["RX_TRIG_SA_PROGR"][:, 1] * (1 / 2.8e6)
        - edr.geo["SPACECRAFT_ALTITUDE"] * 2e3 / c
        + 256 * dt
        - 450e-6
    )

    # rgF1 = np.zeros(edr.data["ZERO_F1"].shape, dtype=np.float32)
    rgF1, psisF1 = pc_clutter(
        edr.data["ZERO_F1"], sim, tshiftF1, edr.ost["DCG_CONFIGURATION_F1"]
    )
    rgF1 += pc_clutter(
        edr.data["MINUS1_F1"],
        sim,
        tshiftF1,
        edr.ost["DCG_CONFIGURATION_F1"],
        psis=psisF1,
    )
    rgF1 += pc_clutter(
        edr.data["PLUS1_F1"],
        sim,
        tshiftF1,
        edr.ost["DCG_CONFIGURATION_F1"],
        psis=psisF1,
    )

    rgF2, psisF2 = pc_clutter(
        edr.data["ZERO_F2"], sim, tshiftF2, edr.ost["DCG_CONFIGURATION_F2"]
    )
    rgF2 += pc_clutter(
        edr.data["MINUS1_F2"],
        sim,
        tshiftF2,
        edr.ost["DCG_CONFIGURATION_F2"],
        psis=psisF2,
    )
    rgF2 += pc_clutter(
        edr.data["PLUS1_F2"],
        sim,
        tshiftF2,
        edr.ost["DCG_CONFIGURATION_F2"],
        psis=psisF2,
    )

    # Normalize to background
    kernel = np.ones(32)
    for i in range(rgF1.shape[1]):
        smooth = np.correlate(rgF1[:, i], kernel, mode="valid")
        ms = np.min(smooth)
        if ms == 0:
            ms = 1
        rgF1[:, i] = rgF1[:, i] / ms

    for i in range(rgF2.shape[1]):
        smooth = np.correlate(rgF2[:, i], kernel, mode="valid")
        ms = np.min(smooth)
        if ms == 0:
            ms = 1
        rgF2[:, i] = rgF2[:, i] / ms

    return rgF1, rgF2, psisF1, psisF2



#!/usr/bin/env python

import argparse
import os
import sys
import re

import numpy as np

import marsis

# To print failure messages and exit
def fail(msg):
    print("\033[91m" + msg + "\033[0m", file=sys.stderr)
    sys.exit(1)


def process(args):
    name = os.path.basename(args.track).replace(".lbl", "")

    edr = marsis.EDR(args.track)

    lat = edr.geo["SUB_SC_LATITUDE"]
    lon = (edr.geo["SUB_SC_LONGITUDE"] + 180) % 360 - 180
    alt = edr.geo["SPACECRAFT_ALTITUDE"]

    bands = np.array([1.8, 3, 4, 5])
    f1_center = bands[edr.ost["DCG_CONFIGURATION_F1"]]
    f2_center = bands[edr.ost["DCG_CONFIGURATION_F2"]]

    if args.method == "none":
        f1, f2 = marsis.plain(edr)
        info = np.dstack((lat, lon, alt, f1_center, f2_center))[0]
        names = "SUB_SC_LATITUDE,SUB_SC_LONGITUDE,SPACECRAFT_ALTITUDE,BAND_F1,BAND_F2"
        fmt = "%.6f,%.6f,%.3f,%.1f,%.1f"
    elif args.method == "campbell":
        f1, f2, f1_rate, f2_rate = marsis.campbell(edr, args.dem)
        info = np.dstack((lat, lon, alt, f1_center, f2_center, f1_rate, f2_rate))[0]
        names = "SUB_SC_LATITUDE,SUB_SC_LONGITUDE,SPACECRAFT_ALTITUDE,BAND_F1,BAND_F2,RATE_F1,RATE_F2"
        fmt = "%.6f,%.6f,%.3f,%.1f,%.1f,%.6e,%.6e"
    elif args.method == "mcmichael":
        f1, f2, f1_psis, f2_psis = marsis.mcmichael(edr, args.sim)
        info = np.dstack(
            (
                lat,
                lon,
                alt,
                f1_center,
                f2_center,
                f1_psis["psi1"],
                f1_psis["psi2"],
                f1_psis["psi3"],
                f2_psis["psi1"],
                f2_psis["psi2"],
                f2_psis["psi3"],
            )
        )[0]
        names = "SUB_SC_LATITUDE,SUB_SC_LONGITUDE,SPACECRAFT_ALTITUDE,BAND_F1,BAND_F2,p1_F1,p2_F1,p3_F1,p1_F2,p2_F2,p3_F2"
        fmt = "%.6f,%.6f,%.3f,%.1f,%.1f,%.6e,%.6e,%.6e,%.6e,%.6e,%.6e"
    with open(args.out + "/" + name + "_nav.csv", "w") as fd:
        fd.write(names + "\n")
        np.savetxt(fd, info, fmt)

    # Write out radargram and nav products
    f1Path = args.out + "/" + name + "_f1.img"
    f2Path = args.out + "/" + name + "_f2.img"

    f1.T.tofile(f1Path)
    f2.T.tofile(f2Path)

    marsis.gen_tiff(f1, f1Path.replace(".img", ".tif"))
    marsis.gen_tiff(f2, f2Path.replace(".img", ".tif"))

    marsis.gen_segy(f1, edr, f1Path.replace(".img", ".sgy"))
    marsis.gen_segy(f2, edr, f2Path.replace(".img", ".sgy"))

    return 0


def cli():
    parser = argparse.ArgumentParser(description="Process MARSIS EDRs")
    parser.add_argument(
        "track", type=str, help="MARSIS EDR file to process (label file)"
    )
    parser.add_argument(
        "method",
        type=str,
        choices=["none", "campbell", "mcmichael"],
        help="Ionosphere compensation method",
    )
    parser.add_argument(
        "-d",
        "--dem",
        type=str,
        help="Global digital elevation model of Mars (reqd. for campbell method)",
        default=None,
    )
    parser.add_argument(
        "-s",
        "--sim",
        type=str,
        help="Surface clutter simulation (reqd. for mcmichael method)",
        default=None,
    )
    parser.add_argument(
        "-o", "--out", type=str, help="Output directory (Default = ./)", default="./"
    )
    parser.add_argument(
        "-n",
        "--num_proc",
        type=int,
        help="Number of processes to spawn (Default = 1)",
        default=1,
    )
    parser.add_argument("-v", "--verbose", action="count", default=0)
    return parser.parse_args()


def check_args(args):
    # Check user's arguments
    # TODO: Check format of track, dem, sim files

    args.out = os.path.abspath(args.out)
    args.track = os.path.abspath(args.track)

    if args.method == "campbell":
        if args.dem is None:
            fail("Global DEM unspecified. Required for campbell ionosphere correction.")
        args.dem = os.path.abspath(args.dem)
        if not os.path.isfile(args.dem):
            fail('Global DEM "%s" cannot be found' % args.dem)

    if args.method == "mcmichael":
        if args.sim is None:
            fail(
                "Surface clutter simulation unspecified. Required for mcmichael ionosphere correction."
            )
        args.sim = os.path.abspath(args.sim)
        if not os.path.isfile(args.sim):
            fail('Surface clutter simulation "%s" cannot be found' % args.sim)

    # Check paths
    if not os.path.isdir(args.out):
        fail('Output directory "%s" cannot be found' % args.out)

    if not os.path.isfile(args.track):
        fail('EDR label file "%s" cannot be found' % args.track)

    if not os.path.isfile(args.track.replace(".lbl", "_f.dat")):
        fail('EDR data file "%s" cannot be found' % args.data)

    if not os.path.isfile(args.track.replace(".lbl", "_g.dat")):
        fail('EDR data file "%s" cannot be found' % args.data)

    if args.verbose:
        print("EDR label file:\t\t%s" % args.track)
        print("EDR data file:\t\t%s" % args.track.replace(".lbl", "_f.dat"))
        print("EDR data file:\t\t%s" % args.track.replace(".lbl", "_g.dat"))
        print("Output directory:\t%s" % args.out)
        if args.dem is not None:
            print("Global DEM:\t\t%s" % args.dem)
        if args.sim is not None:
            print("Clutter simulation:\t%s" % args.sim)

    return args


def main():
    args = cli()
    args = check_args(args)
    process(args)
    return 0


if __name__ == "__main__":
    main()



import numpy as np
from tqdm import tqdm

from .util import refChirp


def pc(DATA, ttrig):
    # TRACE - freq domain data to pulse compress
    # ttrig - trigger time correction

    chirp = refChirp()

    # Zero pad chirp
    chirp = np.append(chirp, np.zeros(DATA.shape[0] - len(chirp)))

    # Shift data to baseband
    fs = 1.4e6
    t = np.arange(len(chirp)) * 1.0 / fs
    bb = np.exp(2 * np.pi * 1j * -0.7e6 * t)
    data = np.fft.ifft(DATA, axis=0)
    data = data * bb[:, np.newaxis]

    # FFT
    CHIRP = np.fft.fft(chirp)
    DATA = np.fft.fft(data, axis=0)

    # Pulse compress
    w = (np.fft.fftfreq(len(chirp), d=1.0 / fs) * 2 * np.pi)[
        :, np.newaxis
    ]  # Angular frequency
    ttrig = ttrig[np.newaxis, :]

    c = 299792458  # speed of light m/s
    pc = np.fft.ifft(
        DATA * np.conj(CHIRP)[:, np.newaxis] * np.exp(-1j * w * ttrig), axis=0
    )
    return np.abs(pc)


def plain(edr):
    c = 299792458
    dt = 1.0 / 1.4e6

    tshiftF1 = (
        edr.anc["RX_TRIG_SA_PROGR"][:, 0] * (1 / 2.8e6)
        - edr.geo["SPACECRAFT_ALTITUDE"] * 2e3 / c
        + 256 * dt
    )
    tshiftF2 = (
        edr.anc["RX_TRIG_SA_PROGR"][:, 1] * (1 / 2.8e6)
        - edr.geo["SPACECRAFT_ALTITUDE"] * 2e3 / c
        + 256 * dt
        - 450e-6
    )

    rgF1 = np.zeros(edr.data["ZERO_F1"].shape, dtype=np.float32)
    rgF1 += pc(edr.data["ZERO_F1"], tshiftF1)
    rgF1 += pc(edr.data["MINUS1_F1"], tshiftF1)
    rgF1 += pc(edr.data["PLUS1_F1"], tshiftF1)

    rgF2 = np.zeros(edr.data["ZERO_F2"].shape, dtype=np.float32)
    rgF2 += pc(edr.data["ZERO_F2"], tshiftF2)
    rgF2 += pc(edr.data["MINUS1_F2"], tshiftF2)
    rgF2 += pc(edr.data["PLUS1_F2"], tshiftF2)

    # Normalize to background
    kernel = np.ones(32)
    for i in range(rgF1.shape[1]):
        smooth = np.correlate(rgF1[:, i], kernel, mode="valid")
        ms = np.min(smooth)
        if ms == 0:
            ms = 1
        rgF1[:, i] = rgF1[:, i] / ms

    for i in range(rgF2.shape[1]):
        smooth = np.correlate(rgF2[:, i], kernel, mode="valid")
        ms = np.min(smooth)
        if ms == 0:
            ms = 1
        rgF2[:, i] = rgF2[:, i] / ms

    return rgF1, rgF2



#!/usr/bin/env python

import argparse
import sys
import os
import re

sys.path.append("/home/mchristo/proj/orbitRadar/marsis_processor/src")

import marsis

# To print failure messages and exit
def fail(msg):
    print("\033[91m" + msg + "\033[0m", file=sys.stderr)
    sys.exit(1)


def cli():
    parser = argparse.ArgumentParser(
        description="Download MARSIS EDRs. Takes list of tracks from a file or stdin"
    )
    parser.add_argument(
        "-t",
        "--tracks",
        type=str,
        help="List of MARSIS EDRs to download (label file)",
        default=None,
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="Directory to save downloaded MARSIS data to (default = ./",
        default="./",
    )
    parser.add_argument("-v", "--verbose", action="count", default=0)
    return parser.parse_args()


def main():
    args = cli()

    if args.tracks is None:
        tracks = sys.stdin.readlines()
    else:
        if not os.path.isfile(args.tracks):
            fail('EDR track list "%s" cannot be found' % args.tracks)
        with open(args.tracks, "r") as fd:
            tracks = fd.readlines()

    # Check paths
    if not os.path.isdir(args.output):
        fail('Output directory "%s" cannot be found' % args.out)

    tracks = [track.strip() for track in tracks]

    # Check track file formatting
    marsis_pattern = re.compile("e_[0-9]{5}_ss3_trk_cmp_m")
    for track in tracks:
        if marsis_pattern.match(track) is None:
            fail('Invalid track "%s"' % track)

    marsis.fetch(tracks, args.output, clobber=True)

    return 0


if __name__ == "__main__":
    main()



"""Utility functions"""

import numpy as np


def arangeT(tstart, tstop, fs):
    """Generate a uniformly sampled timeseries.

    Generate a uniformly sampled timeseries from tstart to tstop
    sampled at frequency fs.

    Parameters
    ----------
    tstart: float
        Starting time
    path: str
        Path to directory where the MARSIS data files will be downloaded to

    Returns
    -------
    seq: array_like of double
        Array containing timeseries values.

    Notes
    -----
    The timeseries generated is a half open interval - tstart is included but
    never tstop.
    """

    tTot = tstop - tstart
    nsamp = int(tTot * fs)
    seq = np.arange(nsamp, dtype=np.double)
    seq = seq / fs + tstart
    return seq


def refChirp(f0=-0.5e6, f1=0.5e6, m=4.0e9, fs=1.4e6):
    """Generate a chirp with given parameters.

    Generate a linear frequency sweep (chirp) with the given frequency
    bound and sweep rate, sampled at a given frequency

    Parameters
    ----------
    f0: float
        Starting frequency of the chirp
    f1: float
        Final frequency of the chirp
    m: float
        Rate of frequency sweep, a positive value is an upsweep and a
        negative value is a downsweep
    fs: float
        Sampling rate of chirp

    Returns
    -------
    chirp: array_like of float
        Tuple containing the sampled chirp amplitude values.

    Notes
    -----
    The time length of the chirp is calculated from the input parameters,
    t = (f1-f0)/m. This function generates a signal with unit amplitude.
    """
    # TODO: Add windowing

    tlen = (f1 - f0) / m
    t = arangeT(0, tlen, fs)  # time vector

    if len(t) == 0:  # In case m is very quick, spit out one sample
        t = np.array([1.0 / fs])

    phi = np.pi * (2 * f0 * t + m * (t ** 2))  # phase
    chirp = -1 * np.sin(phi) + 1j * np.cos(phi)  # chirp
    return chirp



import sys

import numpy as np
import pyproj
import rasterio as rio
import scipy.signal.windows

from .util import refChirp


def lsindx(length):
    # Trace index for least squares
    # Divided by 100 to prevent numerical problems
    return np.arange(length) / 100 + 1


def contrastFit(DATA, ttrig, init=None):
    # Run contrast method over reasonable value and fit curve to it
    # as initial guess for optimization
    m = np.linspace(4e9, 8e9, 100)
    c = np.zeros((len(m), DATA.shape[1]))

    for j in range(DATA.shape[1]):
        for i in range(len(m)):
            c[i, j] = pc_Contrast([m[i]], DATA[:, j], ttrig)

    # Least squares fit
    b = m[np.argmin(c, axis=0)]
    indx = lsindx(len(b))
    A = np.vstack((indx ** 0, indx ** 1, indx ** 2, indx ** 3, indx ** 4, indx ** 5))
    # print(A.T)
    init = scipy.linalg.lstsq(A.T, b)[0]

    return init


def bcOptimize(DATA, ttrig, init=None):
    # Run contrast method over reasonable value and fit curve to it
    # as initial guess for optimization
    m = np.linspace(4e9, 8e9, 100)
    c = np.zeros((len(m), DATA.shape[1]))

    for j in range(DATA.shape[1]):
        for i in range(len(m)):
            c[i, j] = pc_Contrast([m[i]], DATA[:, j], ttrig)

    # Least squares fit
    b = m[np.argmin(c, axis=0)]
    indx = lsindx(len(b))
    A = np.vstack((indx ** 0, indx ** 1, indx ** 2, indx ** 3, indx ** 4, indx ** 5))
    # print(A.T)
    init = scipy.linalg.lstsq(A.T, b)[0]

    # Smooth function SNR based optimization
    res = scipy.optimize.minimize(
        pc_SNR,
        init,
        args=(DATA, ttrig),
        method="Nelder-Mead",
        options={"xatol": 1, "adaptive": True, "maxiter": 50},
    )
    return res.x


def pc_Contrast(b, DATA, ttrig):
    data = pc(b, DATA, ttrig)
    data = np.abs(data)
    contrast = -1 * (np.std(data) / np.mean(data))
    return contrast


def pc_SNR(b, DATA, ttrig):
    np.seterr(all="raise")
    data = pc(b, DATA, ttrig)

    # Calculate summed SNR for the track
    # Take N as the avg of the first 32 samples
    # Take S as the max value of the trace

    data = np.abs(data)
    # Normalize to low echo
    for i in range(data.shape[1]):
        kernel = np.ones(32)
        smooth = np.correlate(data[:, i], kernel, mode="valid")
        ms = np.min(smooth)
        if ms == 0:
            ms = 1
        data[:, i] = data[:, i] / ms
    # N = np.mean(np.abs(data[:32, :]), axis=0)
    # S = np.max(np.abs(data), axis=0)

    # SNR = S/N
    SNR = np.max(data, axis=0)
    sys.stdout.write("SNR=%f\r" % (-1 * np.sum(SNR)))
    return -1 * np.sum(SNR)


def pc(b, DATA, ttrig):
    # b is a vector of floats, coefficents for a polynomial
    # describing chirp rate vs trace index for the chirp
    # Default b for no chirp rate compensation would be [4.0e9]
    # This is not default val to enable plugging this method
    # into minimizers

    # print(DATA.shape, ttrig.shape)

    DATA = np.copy(DATA)

    if len(DATA.shape) == 1:
        DATA = DATA[:, np.newaxis]

    # Base band and filter
    t = np.arange(DATA.shape[0]) * 1.0 / 1.4e6
    bb = np.exp(2 * np.pi * 1j * -0.7e6 * t)
    data = np.fft.ifft(DATA, axis=0)
    data = data * bb[:, np.newaxis]

    # bfilt, afilt = scipy.signal.butter(4, 0.4e6, "low", fs=1.4e6)
    # data = scipy.signal.filtfilt(bfilt, afilt, data, axis=0)

    DATA = np.fft.fft(data, axis=0)

    ntrace = DATA.shape[1]
    indx = lsindx(ntrace)

    # Calculate phase distortion from given coefficents
    m = np.zeros(len(indx), dtype=np.float128)
    for j in range(len(b)):
        m += b[j] * (indx ** j)

    w = np.fft.fftfreq(DATA.shape[0], d=1.0 / 1.4e6) * 2 * np.pi  # Angular frequency

    # Pulse compress frame by frame
    for j in range(ntrace):
        chirp = refChirp(m=m[j])

        # window
        chirp = chirp * scipy.signal.windows.hann(len(chirp))

        # Handle chirp that is too long
        pad = DATA.shape[0] - len(chirp)
        if pad < 0:
            chirp = chirp[0 : DATA.shape[0]]
        else:
            chirp = np.append(chirp, np.zeros(pad))

        CHIRP = np.fft.fft(chirp)

        DATA[:, j] = DATA[:, j] * np.conj(CHIRP) * np.exp(-1j * w * ttrig[j])

    return np.fft.ifft(DATA, axis=0)


def campbell(edr, dem, cacheIono=False, cache="./", contrast=False):
    """Campbell style ionospheric correction.

    Fit a smooth function to quadratic phase distortion caused
    by the ionosphere, where the max SNR of each frame is a used
    as a goodness metric.

    Parameters
    ----------
    edr: EDR
        EDR object containing the EDR data
    path: str
        Path to directory where the MARSIS data files will be downloaded to
    clobber: bool
        Whether to overwrite (re-download) files at path

    Returns
    -------
    files: list of str
        Tuple containing the paths of all files downloaded by this function

    Notes
    -----
    """
    c = 299792458
    dt = 1.0 / 1.4e6

    # RX window shift
    tshiftF1 = (
        edr.anc["RX_TRIG_SA_PROGR"][:, 0] * (1 / 2.8e6)
        - edr.geo["SPACECRAFT_ALTITUDE"] * 2e3 / c
        + 256 * dt
    )
    tshiftF2 = (
        (edr.anc["RX_TRIG_SA_PROGR"][:, 1]) * (1 / 2.8e6)
        - edr.geo["SPACECRAFT_ALTITUDE"] * 2e3 / c
        + 256 * dt
        - 450e-6  # Delay from xmit F1 to xmit F2
    )

    # Find when freq switch is (if any)
    dcgF1 = edr.ost["DCG_CONFIGURATION_F1"].to_numpy()
    dcgF2 = edr.ost["DCG_CONFIGURATION_F2"].to_numpy()

    switchF1 = np.where(dcgF1[:-1] != dcgF1[1:])[0] + 1
    switchF2 = np.where(dcgF1[:-1] != dcgF1[1:])[0] + 1

    switchF1 = np.append(0, switchF1)
    switchF2 = np.append(0, switchF2)

    switchF1 = np.append(switchF1, len(dcgF1))
    switchF2 = np.append(switchF2, len(dcgF2))

    f1 = edr.data["ZERO_F1"]
    f2 = edr.data["ZERO_F2"]

    # TODO: Add cache and reading from cache for ionosphere coefficents

    # cacheFile = "%s/%s_coeffs.txt" % (cache, edr.lbld["PRODUCT_ID"].lower())

    # if cacheIono:
    #    if(os.path.isfile(cacheFile)):
    #        fd = open(cacheFile, "r")
    #        coeffs = fd.read().split("\n")
    #        fd.close()
    #        coeffDict = {}
    #        for line in coeffs:
    #            line = line.split("=")
    #            coeffDict[line.]"""

    # Ionosphere corrections for each continuious chunk of band in each frequency
    # could probably do some f1-f2 stitching but that would take more thinking
    bF1 = []
    for i in range(len(switchF1) - 1):
        bF1.append(
            bcOptimize(
                f1[:, switchF1[i] : switchF1[i + 1]],
                tshiftF1[switchF1[i] : switchF1[i + 1]],
            )
        )

    bF2 = []
    for i in range(len(switchF2) - 1):
        bF2.append(
            bcOptimize(
                f2[:, switchF2[i] : switchF2[i + 1]],
                tshiftF2[switchF2[i] : switchF2[i + 1]],
            )
        )

    # Write coeffs to file
    # if cacheIono:
    #    fd = open(cacheFile, "w")
    #    elems = "%e, " * len(bF1B1)
    #    bF1B1Line = "bF1B1 = [" + elems[:-2] + "]\n"
    #    elems = "%e, " * len(bF1B2)
    #    bF1B2Line = "bF1B2 = [" + elems[:-2] + "]\n"
    #    elems = "%e, " * len(bF2B1)
    #    bF2B1Line = "bF2B1 = [" + elems[:-2] + "]\n"
    #    elems = "%e, " * len(bF2B2)
    #    bF2B2Line = "bF2B2 = [" + elems[:-2] + "]\n"

    #    fd.write(bF1B1Line % tuple(bF1B1))
    #    fd.write(bF1B2Line % tuple(bF1B2))
    #    fd.write(bF2B1Line % tuple(bF2B1))
    #    fd.write(bF2B2Line % tuple(bF2B2))
    #    fd.close()

    # Make multilook images for f1 and f2
    rgF1 = np.zeros_like(f1, dtype=np.float32)
    rateF1 = np.zeros(f1.shape[1], dtype=np.float32)
    for i in range(len(switchF1) - 1):
        for j in ["MINUS1_F1", "ZERO_F1", "PLUS1_F1"]:
            rgF1[:, switchF1[i] : switchF1[i + 1]] += np.abs(
                pc(
                    bF1[i],
                    edr.data[j][:, switchF1[i] : switchF1[i + 1]],
                    tshiftF1[switchF1[i] : switchF1[i + 1]],
                )
            )
            ntrace = switchF1[i + 1] - switchF1[i]
            indx = lsindx(ntrace)

            m = np.zeros(len(indx), dtype=np.float128)
            for k in range(len(bF1[i])):
                m += bF1[i][k] * (indx ** k)

            rateF1[switchF1[i] : switchF1[i + 1]] = m

    rgF2 = np.zeros_like(f2, dtype=np.float32)
    rateF2 = np.zeros(f1.shape[1], dtype=np.float32)
    for i in range(len(switchF2) - 1):
        for j in ["MINUS1_F2", "ZERO_F2", "PLUS1_F2"]:
            rgF2[:, switchF2[i] : switchF2[i + 1]] += np.abs(
                pc(
                    bF2[i],
                    edr.data[j][:, switchF2[i] : switchF2[i + 1]],
                    tshiftF2[switchF2[i] : switchF2[i + 1]],
                )
            )
            ntrace = switchF2[i + 1] - switchF2[i]
            indx = lsindx(ntrace)

            m = np.zeros(len(indx), dtype=np.float128)
            for k in range(len(bF2[i])):
                m += bF2[i][k] * (indx ** k)

            rateF2[switchF2[i] : switchF2[i + 1]] = m

    # Normalize to background
    kernel = np.ones(32)
    for i in range(rgF1.shape[1]):
        smooth = np.correlate(rgF1[:, i], kernel, mode="valid")
        ms = np.min(smooth)
        if ms == 0:
            ms = 1
        rgF1[:, i] = rgF1[:, i] / ms

    for i in range(rgF2.shape[1]):
        smooth = np.correlate(rgF2[:, i], kernel, mode="valid")
        ms = np.min(smooth)
        if ms == 0:
            ms = 1
        rgF2[:, i] = rgF2[:, i] / ms

    # Correct to MOLA surface
    # srfF1 = np.argmax(rgF1 > thld, axis=0)
    # srfF2 = np.argmax(rgF2 > thld, axis=0)

    # srfF1 = np.argmax(rgF1, axis=0)
    # srfF2 = np.argmax(rgF2, axis=0)

    srfF1 = np.argmax(10 * np.log(rgF1) > -15, axis=0)
    srfF2 = np.argmax(10 * np.log(rgF2) > -10, axis=0)

    # If no surface found just set surface to halfway, it will be a junk
    # track anyway
    if np.all(srfF1 == 0):
        srfF1[:] = 256

    if np.all(srfF2 == 0):
        srfF2[:] = 256

    x = np.arange(len(srfF1))
    if len(x[srfF1 != 0]) != 0:
        # Just skipping bad tracks for now
        srfF1 = np.interp(x, x[srfF1 != 0], srfF1[srfF1 != 0])
        srfF2 = np.interp(x, x[srfF2 != 0], srfF2[srfF2 != 0])

    # MOLA Surface
    # xyzcrs = "+proj=geocent +R=3396000 +lat_0=0 +lon_0=180 +type=crs"
    # xyzcrs = "+proj=geocent +a=3396000 +no_defs"
    # xyzcrs = "+proj=geocent +a=3396190 +b=3376200 +no_defs"
    # xyzcrs = "+proj=geocent +lat_ts=0 +lat_0=0 +lon_0=0 +x_0=0 +y_0=0 +R=3396000 +units=m +no_defs"
    # xyzcrs = "+proj=geocent +a=3396190 +b=3376200 +no_defs"
    # xyzcrs = "+proj=geocent +R=3396000 +no_defs"
    llacrs = 'GEOGCS["unknown",DATUM["unknown",SPHEROID["unknown",3396000,0]],PRIMEM["Reference meridian",0],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AXIS["Latitude",NORTH],AXIS["Longitude",EAST]]'

    xyz = edr.geo["TARGET_SC_POSITION_VECTOR"] * 1e3

    lat = edr.geo["SUB_SC_LATITUDE"]
    lon = (edr.geo["SUB_SC_LONGITUDE"] + 180) % 360 - 180
    # alt = edr.geo["SPACECRAFT_ALTITUDE"]
    # lla = np.dstack((lon, lat, alt))
    dem = rio.open(dem, "r")

    # print(dem.crs, xyzcrs)

    # demX, demY, demZ = pyproj.transform(
    #    llacrs, dem.crs, lla[:, 0], lla[:, 1], lla[:, 2]
    # )

    iy, ix = dem.index(lon, lat)
    ix = np.array(ix)
    iy = np.array(iy)

    # Temp fix mola meters/pix issue
    ix[ix > dem.width - 1] = dem.width - 1
    ix[ix < 0] = 0

    mola = dem.read(1)[iy, ix] + 3396000

    dem.close()

    angle = np.abs(np.arctan(xyz[:, 2] / np.sqrt(xyz[:, 0] ** 2 + xyz[:, 1] ** 2)))

    a = 3396190
    b = 3376200

    marsR = (a * b) / np.sqrt(
        (a ** 2) * (np.sin(angle) ** 2) + (b ** 2) * (np.cos(angle) ** 2)
    )

    c = 299792458
    molasample = 256 + (((marsR - mola) * 2 / c) * 1.4e6).astype(np.int32)

    dsF1 = (molasample - srfF1).astype(np.int32)
    dsF2 = (molasample - srfF2).astype(np.int32)

    for i in range(rgF1.shape[1]):
        rgF1[:, i] = np.roll(rgF1[:, i], dsF1[i])
        rgF2[:, i] = np.roll(rgF2[:, i], dsF2[i])

    return rgF1, rgF2, rateF1, rateF2



import struct

import numpy as np
import pandas as pd
import PIL as pil
import matplotlib.pyplot as plt


def gen_tiff(rg, name):
    # Log scaling
    rg = np.log(np.power(rg, 2))
    rg -= rg.min()
    rg *= 255 / rg.max()
    rg = rg.astype(np.uint8)
    img = pil.Image.fromarray(rg)
    img.save(name)


def gen_segy(rg, edr, name):
    lat = np.copy(edr.geo["SUB_SC_LATITUDE"])
    lon = np.copy(edr.geo["SUB_SC_LONGITUDE"])

    # Convert 0-360 to -180-180
    for i in range(len(lon)):
        if lon[i] > 180:
            lon[i] = lon[i] - 360

        lon[i] = lon[i]

    # Multiply by 10000 to store as ints w/ 3 decimal digit precision in segy
    lat *= 10000
    lon *= 10000

    # Load data
    rg = np.transpose(rg).copy(order="C")  # This makes the traces C-contiguous
    spt = rg.shape[1]
    ntrace = rg.shape[0]

    # Build Segy 2.0 format files
    # https://seg.org/Portals/0/SEG/News%20and%20Resources/Technical%20Standards/seg_y_rev2_0-mar2017.pdf
    with open(name, "wb") as f:
        ### Write 3200 byte text header
        f.write(build_txthead(name.split("/")[-1].split(".")[0]))

        ### Write 400 byte binary header
        f.write(build_binhead(spt, ntrace))

        ### Write data traces
        for i in range(ntrace):
            f.write(build_trchead(spt, i + 1, int(lat[i]), int(lon[i])))
            f.write(rg[i, :].astype(">f", order="C").tobytes(order="C"))

    return 0


def build_txthead(track):
    txthead = b""

    # (0) Track number string
    track = track[0:10]
    txthead += track.encode(encoding="ASCII", errors="strict")

    # (10) Rest of text header, 3200-10 = 3190 bytes
    for i in range(398):
        txthead += struct.pack("Q", 0)

    txthead += struct.pack("I", 0)
    txthead += struct.pack("H", 0)

    return txthead


def build_binhead(spt, ntrace):
    ### Build binary header
    binhead = b""

    # (0) Job identification number
    binhead += struct.pack(">I", 1)
    # (4) Line number
    binhead += struct.pack(">I", 1)
    # (8) Reel number
    binhead += struct.pack(">I", 1)
    # (12) Traces per ensemble
    binhead += struct.pack(">H", 1)
    # (14) Auxiliary traces per ensemble
    binhead += struct.pack(">H", 0)
    # (16) Sample interval
    binhead += struct.pack(">H", 375)
    # (18) Field recording sample interval
    binhead += struct.pack(">H", 375)
    # (20) Samples per trace
    binhead += struct.pack(">H", spt)
    # (22) Field recording samples per trace
    binhead += struct.pack(">H", spt)
    # (24) Data format code
    binhead += struct.pack(">H", 5)
    # (26) Ensemble fold
    binhead += struct.pack(">H", 1)
    # (28) Trace sorting code
    binhead += struct.pack(">H", 4)
    # (30) Vertical sum code
    binhead += struct.pack(">H", 1)
    # (32) Sweep freq start
    binhead += struct.pack(">H", 0)
    # (34) Sweep freq end
    binhead += struct.pack(">H", 0)
    # (36) Sweep length
    binhead += struct.pack(">H", 0)
    # (38) Sweep type
    binhead += struct.pack(">H", 0)
    # (40) Trace number of sweep channel
    binhead += struct.pack(">H", 0)
    # (42) Sweep trace taper length at start
    binhead += struct.pack(">H", 0)
    # (44) Sweep trace taper length at end
    binhead += struct.pack(">H", 0)
    # (46) Taper type
    binhead += struct.pack(">H", 0)
    # (48) Correlated data traces, 2 is yes
    binhead += struct.pack(">H", 2)
    # (50) Binary gain recovered, 2 is no
    binhead += struct.pack(">H", 1)
    # (52) Amplitude recovery method, 1 is none
    binhead += struct.pack(">H", 4)
    # (54) Measurement system, 1 is meters
    binhead += struct.pack(">H", 1)
    # (56) Implulse signal polarity
    binhead += struct.pack(">H", 1)
    # (58) Vibratory polarity code
    binhead += struct.pack(">H", 0)
    # (60) Extended number of traces per ensemble
    binhead += struct.pack(">I", 0)
    # (64) Extended number of auxiliary traces per ensemble
    binhead += struct.pack(">I", 0)
    # (68) Extended samples per trace
    binhead += struct.pack(">I", 0)
    # (72) Extended sample interval
    binhead += struct.pack(">d", 37.5)
    # (80) Extended field recording sample interval
    binhead += struct.pack(">d", 0)
    # (88) Extended field recording samples per trace
    binhead += struct.pack(">I", 0)
    # (92) Extended ensemble fold
    binhead += struct.pack(">I", 0)
    # (96) An integer constant - see format doc
    binhead += struct.pack(">I", 16909060)
    # (100) 200 unassigned bytes
    for i in range(25):
        binhead += struct.pack("Q", 0)
    # (300) Major SEG-Y format revision number
    binhead += struct.pack(">B", 2)
    # (301) Minor SEG-Y format revision number
    binhead += struct.pack(">B", 0)
    # (302) Fixed length trace flag, 1 is yes - see format doc
    binhead += struct.pack(">H", 1)
    # (304) Number of extended text file headers
    binhead += struct.pack(">H", 0)
    # (306) Number of additional trace headers
    binhead += struct.pack(">I", 0)
    # (310) Time basis code, 4 is UTC
    binhead += struct.pack(">H", 4)
    # (312) Number of traces in file
    binhead += struct.pack(">Q", ntrace)
    # (320) Byte offset of first trace
    binhead += struct.pack(">Q", 0)
    # (328) Number of data trailer stanza records
    binhead += struct.pack(">I", 0)
    # (332) 68 unassigned bytes
    for i in range(17):
        binhead += struct.pack(">I", 0)

    return binhead


def build_trchead(spt, i, lat, lon):
    # i - trace number
    trchead = b""
    # (0) Trace sequence number within line
    trchead += struct.pack(">I", i)
    # (4) Trace sequence number within file - starts at 1
    trchead += struct.pack(">I", i)
    # (8) Original field record number
    trchead += struct.pack(">I", i)
    # (12) Trace number in original field record
    trchead += struct.pack(">I", i)
    # (16) Energy source point number
    trchead += struct.pack(">I", 0)
    # (20) Ensemble number
    trchead += struct.pack(">I", i)
    # (24) Trace number within ensemble
    trchead += struct.pack(">I", 0)
    # (28) Trace ID code, 1 is time domain
    trchead += struct.pack(">H", 1)
    # (30) Number of vertically summed traces
    trchead += struct.pack(">H", 0)
    # (32) Number of horizontally stacked traces (stacking??)
    trchead += struct.pack(">H", 0)
    # (34) Data use, 1 is production
    trchead += struct.pack(">H", 1)
    # (36) Distance from center of source point to center of rx group
    trchead += struct.pack(">I", 0)
    # (40) Elevation of rx group
    trchead += struct.pack(">I", 0)
    # (44) Surface elevation at source location
    trchead += struct.pack(">I", 0)
    # (48) Source depth below surface
    trchead += struct.pack(">I", 0)
    # (52) Seismic datum elevation at rx group
    trchead += struct.pack(">I", 0)
    # (56) Seismic datum elevation at source
    trchead += struct.pack(">I", 0)
    # (60) Water column height at source location
    trchead += struct.pack(">I", 0)
    # (64) Water column height at rx group location
    trchead += struct.pack(">I", 0)
    # (68) Scalar to be applied to last seven fields for real val - see format doc
    trchead += struct.pack(">H", 1)
    # (70) Scalar to be applied to four following fields for real val - see format doc
    trchead += struct.pack(">h", -10000)
    # (72) Source X coord
    trchead += struct.pack(">I", 0)
    # (76) Source Y coord
    trchead += struct.pack(">I", 0)
    # (80) rx group X coord
    trchead += struct.pack(">I", 0)
    # (84) rx group Y coord
    trchead += struct.pack(">I", 0)
    # (88) Coordinate units, 3 is decimal degrees
    trchead += struct.pack(">H", 3)
    # (90) Weathering velocity
    trchead += struct.pack(">H", 0)
    # (92) Subweathering velocity
    trchead += struct.pack(">H", 0)
    # (94) Uphole time at source in ms
    trchead += struct.pack(">H", 0)
    # (96) Uphole time at group in ms
    trchead += struct.pack(">H", 0)
    # (98) Source static correction in ms
    trchead += struct.pack(">H", 0)
    # (100) Group static correction in ms
    trchead += struct.pack(">H", 0)
    # (102) Total static correction applied in ms
    trchead += struct.pack(">H", 0)
    # (104) Lag time A
    trchead += struct.pack(">H", 0)
    # (106) Lag time B
    trchead += struct.pack(">H", 0)
    # (108) Delay recording time
    trchead += struct.pack(">H", 0)
    # (110) Mute time start in ms
    trchead += struct.pack(">H", 0)
    # (112) Mute time stop in ms
    trchead += struct.pack(">H", 0)
    # (114) Samples in this trace
    trchead += struct.pack(">H", spt)
    # (116) Sample interval for this trace
    trchead += struct.pack(">H", 0)
    # (118) Gain type of field instruments
    trchead += struct.pack(">H", 0)
    # (120) Instrument gain constant
    trchead += struct.pack(">H", 0)
    # (122) Instrument early or initial gain
    trchead += struct.pack(">H", 0)
    # (124) Correlated, 2 is yes
    trchead += struct.pack(">H", 2)
    # (126) Sweep freq start Hz
    trchead += struct.pack(">H", 0)
    # (128) Sweep freq end Hz
    trchead += struct.pack(">H", 0)
    # (130) Sweep length ms
    trchead += struct.pack(">H", 0)
    # (132) Sweep type, 1 is linear
    trchead += struct.pack(">H", 1)
    # (134) Sweep trace taper length start ms
    trchead += struct.pack(">H", 0)
    # (136) Sweep trace taper length end ms
    trchead += struct.pack(">H", 0)
    # (138) Taper type
    trchead += struct.pack(">H", 0)
    # (140) Alias filter freq Hz
    trchead += struct.pack(">H", 0)
    # (142) Alias filter slope
    trchead += struct.pack(">H", 0)
    # (144) Notch filter freq Hz
    trchead += struct.pack(">H", 0)
    # (146) Notch filter flope
    trchead += struct.pack(">H", 0)
    # (148) Low-cut freq Hz
    trchead += struct.pack(">H", 0)
    # (150) Hi-cut freq Hz
    trchead += struct.pack(">H", 0)
    # (152) Low-cut slope
    trchead += struct.pack(">H", 0)
    # (154) Hi-cut slope
    trchead += struct.pack(">H", 0)
    # (156) Year data recorded
    trchead += struct.pack(">H", 0)
    # (158) Day of year data recorded
    trchead += struct.pack(">H", 0)
    # (160) Hour of day
    trchead += struct.pack(">H", 0)
    # (162) Minute of hour
    trchead += struct.pack(">H", 0)
    # (164) Second of minute
    trchead += struct.pack(">H", 0)
    # (166) Time basis code
    trchead += struct.pack(">H", 0)
    # (168) Trace weighting factor
    trchead += struct.pack(">H", 0)
    # (170) Geophone group number of roll switch position one
    trchead += struct.pack(">H", 0)
    # (172) Geophone group number of trace number one within original field record
    trchead += struct.pack(">H", 0)
    # (174) Geophone group number of last trace within original field record
    trchead += struct.pack(">H", 0)
    # (176) Gap size
    trchead += struct.pack(">H", 0)
    # (178) Over travel assoc with taper at beginning of line
    trchead += struct.pack(">H", 0)
    # (180) X coordinate of ensemble position of this trace
    trchead += struct.pack(">i", lon)
    # (184)Y coordinate of ensemble position of this trace
    trchead += struct.pack(">i", lat)
    # (188) Line number (inline  number)
    trchead += struct.pack(">I", 1)
    # (192) Ensemble number (crossline number)
    trchead += struct.pack(">I", i)
    # (196) Shotpoint number
    trchead += struct.pack(">I", 0)
    # (200) Scalar to be applied to shotpoint number
    trchead += struct.pack(">H", 0)
    # (202) Trace units, 0 is unknown
    trchead += struct.pack(">H", 0)
    # (204) Transduction constant
    trchead += struct.pack(">I", 0)
    trchead += struct.pack(">H", 0)
    # (210) Transduction units
    trchead += struct.pack(">H", 0)
    # (212) Device/Trace identifier
    trchead += struct.pack(">H", 0)
    # (214) Scalar to be applied to uphole, static, and group corrections
    trchead += struct.pack(">H", 0)
    # (216) Source type/orientation
    trchead += struct.pack(">H", 0)
    # (218) Source energy direction wrt orientation
    trchead += struct.pack(">H", 0)
    trchead += struct.pack(">H", 0)
    trchead += struct.pack(">H", 0)
    # (224) Source measurement
    trchead += struct.pack(">I", 0)
    trchead += struct.pack(">H", 0)
    # (230) Source measurement unit
    trchead += struct.pack(">H", 0)
    # (232) Zero
    trchead += struct.pack(">Q", 0)

    return trchead
